/*
 * Copyright (C) 2020 - 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <nil-cairo.h>
#include <nil-cga.h>
#include <nil-pctext.h>

#include "cons_buf.h"
#include "pctextmode.h"


cairo_bool_t
abydos_console_has_blink(Console *cons){
    int size = cons_buf_width(cons) * cons_buf_height(cons);
    const ConsoleCell *p = cons_buf_buffer(cons);
    while(size>0){
        if(p->attr & 0x80)
            return 1;
        --size;
        ++p;
    }
    return 0;
}


cairo_surface_t *
abydos_image_surface_from_console(Console *cons, AbydosConsoleState state, int letter_spacing)
{
    const ConsoleCell *src = cons_buf_buffer(cons);
    int width = cons_buf_width(cons);
    int height = cons_buf_height(cons);
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_RGB24,width*letter_spacing,height*16);
    nile_t ni;
    uint32_t palette[16];
    int y;
    nile_init(&ni);
    nil_cairo_from_surface(&ni, surface);
    nil_cga_copy_palette_colors((uint8_t*)palette, ni.bpc, ni.format, 16);
    for(y=0;y<height;++y){
        int x;
        for(x=0;x<width;++x){
            ConsoleCell cell = *src++;
            uint32_t bg;
            uint32_t fg;
            if(S_ICE==state){
                bg = palette[(cell.attr>>4)&15];
            }else{
                bg = palette[(cell.attr>>4)&7];
            }
            if(!(S_OFF==state && (cell.attr&0x80))){
                fg = palette[(cell.attr)&15];
                nil_pctext_draw_character(&ni, x*letter_spacing, y*16, cell.ch, bg, fg, letter_spacing);
            }else{
                nil_pctext_draw_character(&ni, x*letter_spacing, y*16, ' ', bg, bg, letter_spacing);
            }
        }
    }
    nile_done(&ni);
    cairo_surface_mark_dirty(surface);
    return surface;
}
