#ifndef doc_layer_h
#define doc_layer_h

#include "blend_mode.h"
#include "cel_list.h"
//#include "layer_list.h"
#include "with_user_data.h"

namespace doc{
    class LayerFolder;
    class Sprite;

    enum class LayerFlags{
        Visible    = 1,
        Editable   = 2,
        LockMove   = 4,
        Background = 8,
        Continuous = 16,
        BlackgroundLayerFlags = LockMove | Background
    };

    class Layer : public WithUserData{
    private:
        Sprite *_sprite;
        LayerFolder *_parent;
        LayerFlags _flags;
    protected:
        Layer(Sprite* s){ _sprite = s; }
    public:
        //virtual ~Layer();

        std::string name() const { return ""; }
        void setName(const std::string &name) {}

        Sprite *sprite() const { return _sprite; }
        LayerFolder *parent() const { return _parent; }
        void setParent(LayerFolder *lf) { _parent = lf; }

        virtual bool isImage() const { return false; };
        virtual bool isFolder() const { return false; };

        LayerFlags flags() const { return _flags; }
        void setFlags(LayerFlags f) { _flags = f; }

        virtual Cel *cel(int frame) const { return NULL; };
        virtual void getCels(CelList &cels) const = 0;
    };

    class LayerImage : public Layer{
    private:
        BlendMode _blendmode;
        int _opacity;
        CelList _cels;
    public:
        explicit LayerImage(Sprite *s):Layer(s){}
        //virtual ~LayerImage();

        virtual bool isImage() const { return true; }

        BlendMode blendMode() const { return _blendmode; }
        void setBlendMode(BlendMode bm) { _blendmode = bm; }

        int opacity() const { return _opacity; }
        void setOpacity(int o) { _opacity = o; }

        void addCel(Cel*);

        Cel* cel(int frame) const;
        void getCels(CelList &cels) const;
        void displaceFrames(int from, int delta);

        //virtual int getMemSize() const;
    };

    typedef std::vector<Layer*> LayerList;
    typedef LayerList::iterator LayerIterator;
    typedef LayerList::const_iterator LayerConstIterator;

    class LayerFolder : public Layer{
    private:
        LayerList _layers;
    public:
        explicit LayerFolder(Sprite* s):Layer(s){};
        virtual ~LayerFolder();

        virtual bool isFolder() const { return true; }

        const LayerList &getLayersList() { return _layers; }
        LayerIterator getLayerBegin() { return _layers.begin(); }
        //LayerIterator getLayerEnd() { return _layer.begin; }
        LayerConstIterator getLayerBegin() const { return _layers.begin(); }
        LayerConstIterator getLayerEnd() const { return _layers.end(); }

        void addLayer(Layer*);
        void removeLayer(Layer*);
        //void stackLayer(Layer*, Layer* after);

        void getCels(CelList&) const;
        void displaceFrames(int from, int delta);
    };
}

#endif /* layer_h */
