/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <endian.h>
#include <string.h>                     /* memcpy() */

#include "cgmread.h"
#include "elements.h"
#include "loa-matrix.h"


cgm_read_t *cgm_read_init(cgm_read_t *r){
    r->data = r->end = NULL;

    r->int_precision = r->index_precision = 2;
    r->real_size = r->vdc_real_size = 4;
    r->color_index_precision = 1;
    r->color_precision = 1;
    r->vdc_int_precision = 2;
    r->real_type = r->vdc_real_type = RT_FIXED; /* RP_FLOAT; */

    r->vdc_type = VDC_INT;

    r->eDeviceViewPortMode = DVPM_FRACTION;

    return r;
}


void cgm_read_set(cgm_read_t *r, const uint8_t *data, size_t len){
    r->data = data;
    r->end = data + len;
}


#if 0
void CGM::ImplGetRectangleNS( FloatRect& rFloatRect )
{
    if ( pElement.eVDCType == VDC_INTEGER )
    {
        rFloatRect.Left = ImplGetI( pElement.nVDCIntegerPrecision );
        rFloatRect.Bottom = ImplGetI( pElement.nVDCIntegerPrecision );
        rFloatRect.Right = ImplGetI( pElement.nVDCIntegerPrecision );
        rFloatRect.Top = ImplGetI( pElement.nVDCIntegerPrecision );
    }
    else // ->floating points
    {
        rFloatRect.Left = ImplGetFloat( pElement.eVDCRealPrecision, pElement.nVDCRealSize );
        rFloatRect.Bottom = ImplGetFloat( pElement.eVDCRealPrecision, pElement.nVDCRealSize );
        rFloatRect.Right = ImplGetFloat( pElement.eVDCRealPrecision, pElement.nVDCRealSize );
        rFloatRect.Top = ImplGetFloat( pElement.eVDCRealPrecision, pElement.nVDCRealSize );
    }
}
#endif


#if 1
uint32_t r_get_color(cgm_read_t *r, const CGMElements *e, int direct)
{
    // the background color is always a direct color

    uint32_t  nTmp;
    if ( ( e->eColorSelectionMode == CSM_DIRECT ) || direct )
    {
        uint32_t      nColor = cgm_get_byte( r_get_color_channel(r), 1 );
        uint32_t      nDiff = e->nColorValueExtent[ 3 ] - e->nColorValueExtent[ 0 ] + 1;

        if ( !nDiff )
            nDiff++;
        nColor = ( ( nColor - e->nColorValueExtent[ 0 ] ) << 8 ) / nDiff;
        nTmp = nColor << 16 & 0xff0000;

        nColor = cgm_get_byte( r_get_color_channel(r), 1 );
        nDiff = e->nColorValueExtent[ 4 ] - e->nColorValueExtent[ 1 ] + 1;
        if ( !nDiff )
            nDiff++;
        nColor = ( ( nColor - e->nColorValueExtent[ 1 ] ) << 8 ) / nDiff;
        nTmp |= nColor << 8 & 0xff00;

        nColor = cgm_get_byte( r_get_color_channel(r), 1 );
        nDiff = e->nColorValueExtent[ 5 ] - e->nColorValueExtent[ 2 ] + 1;
        if ( !nDiff )
            nDiff++;
        nColor = ( ( nColor - e->nColorValueExtent[ 2 ] ) << 8 ) / nDiff;
        nTmp |= nColor & 0xff;
    }
    else
    {
        //int nIndex = r_get_uint_x(r, r->color_index_precision );
        int nIndex = r_get_color_index(r);
        nTmp = e->aColorTable[ nIndex & 0xff ] ;
    }
    return nTmp;
}
#endif


void r_get_rect(cgm_read_t *r, rect_t *rect){
    loa_point_t pos2;
    rect->pos.x = r_get_VDC(r);
    pos2.y = r_get_VDC(r);
    pos2.x = r_get_VDC(r);
    rect->pos.y = r_get_VDC(r);
    rect->size.x = pos2.x - rect->pos.x;
    rect->size.y = pos2.y - rect->pos.y;
}

/* --- Basic types --- */

void r_get_DIRECTCOLR(cgm_read_t *r, uint8_t *col)
{
}


void r_get_K(cgm_read_t *r, uint8_t *col)
{
}


void r_get_P(cgm_read_t *r, loa_point_t *p)
{
    p->x = r_get_VDC(r);
    p->y = r_get_VDC(r);
}


double r_get_V(cgm_read_t *r, SpecMode m)
{
    switch(m){
    case SM_ABSOLUTE:
        return r_get_VDC(r);
    case SM_SCALED:
        return r_get_R(r) * 25;
    //case SM_FRACTIONAL:
    //case SM_METRIC:
    }
    /* not reached */
    return 0;
}


double r_get_VC(cgm_read_t *r)
{
    switch(r->eDeviceViewPortMode){
    case DVPM_FRACTION:
        return r_get_R(r);
    case DVPM_METRIC:
    case DVPM_DEVICE:
        return r_get_I(r);
    }
    /* not reached */
    return 0;
}


double r_get_VP(cgm_read_t *r, loa_point_t *p)
{
}
