/*
 * Copyright (C) 2021  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#define PY_SSIZE_T_CLEAN
#include <debug.h>
#include <Python.h>

#include "aceunpack.h"

#ifndef MIN
# define MIN(a,b) ((a)<(b)?(a):(b))
#endif


#if 0
static PyObject *
_test(PyObject *self, PyObject *args)
{
    DEBUG_p("begin");
    Py_RETURN_NONE;
}
#endif


static PyObject *
_create_c_module(void)
{
    static PyMethodDef c_methods[] = {
        //{"test", _test, METH_VARARGS, NULL},
        {NULL, NULL, 0, NULL}
    };
    static PyModuleDef c_module = {
        PyModuleDef_HEAD_INIT, "c", NULL, -1, c_methods,
        NULL, NULL, NULL, NULL
    };
    return PyModule_Create(&c_module);
}

struct _aceunpack_t{
    PyObject *archive;
    PyObject *member;
    int idx;
    PyObject *data;
    off_t pos;
};


#include "aceunpack.py.h"


aceunpack_t *
aceunpack_new(const char *filename)
{
    aceunpack_t *au = malloc(sizeof(aceunpack_t));
    PyObject *mod;
    PyObject *dict;

    //PyImport_AppendInittab("c", _create_c_module);
    Py_Initialize();
    mod = _create_c_module();
    dict = PyModule_GetDict(mod);
    PyRun_String(aceunpack_py, Py_file_input, dict, dict);
    au->archive = PyObject_CallMethod(mod, "AceArchive", "s", filename);
    /* FIXME!! check for errors somehow */
    Py_DECREF(dict);
    Py_DECREF(mod);

    au->member = NULL;
    au->idx = 0;
    au->data = NULL;
    return au;
}


void
aceunpack_free(aceunpack_t *au)
{
    if(au->member)
        Py_DECREF(au->member);
    if(au->data)
        Py_DECREF(au->data);
    PyObject_CallMethod(au->archive, "close", "");
    Py_DECREF(au->archive);
    Py_Finalize();
    free(au);
}



int
aceunpack_next(aceunpack_t *au)
{
    if(au->member)
        Py_DECREF(au->member);
    if(au->data){
        Py_DECREF(au->data);
        au->data = NULL;
    }
    DEBUG_p("looking up member %d",au->idx);
    au->member = PyObject_CallMethod(au->archive, "getmember", "i", au->idx++);
    return au->member ? 0 : -1;
}


const char *
aceunpack_get_name(aceunpack_t *au)
{
    const char *name;
    PyObject *o = PyObject_GetAttrString(au->member, "filename");
    name = PyBytes_AS_STRING(o);
    Py_DECREF(o);
    return name;
}


ssize_t
aceunpack_get_size(aceunpack_t *au)
{
    ssize_t size;
    PyObject *o = PyObject_GetAttrString(au->member, "size");
    size = PyLong_AsLong(o);
    Py_DECREF(o);
    return size;
}


ssize_t
aceunpack_read(aceunpack_t *au, void *buf, size_t bytes)
{
    char *data;
    size_t size;
    if(!au->data){
        au->data = PyObject_CallMethod(au->archive, "read", "O", au->member);
        au->pos = 0;
    }
    data = PyBytes_AS_STRING(au->data);
    size = PyBytes_GET_SIZE(au->data);
    bytes = MIN(bytes, size - au->pos);
    if(bytes <= 0)
        return -1;
    memcpy(buf, data + au->pos, bytes);
    au->pos += bytes;
    return bytes;
}
