/*
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This file is public domain.
 */
#ifndef loa_abstract_abydos_h
#define loa_abstract_abydos_h

#include <abydos.h>
#include <cairo.h>

#include "loa-abstract.h"

struct _loa_abstract_t{
    int refcount;
    enum{
        ABST_EMPTY,
        ABST_ABYDOS,
        ABST_PATTERN,
        ABST_SURFACE
    }type;
    int width;
    int height;
    union{
        abydos_t *ar;
        cairo_pattern_t *pattern;
        cairo_surface_t *surface;
    };
};

loa_abstract_t *loa_abstract_abydos_create(void *userdata, int width, int height);
loa_abstract_t *loa_abstract_abydos_from_abydos(abydos_t*);
void loa_abstract_abydos_render(double x, double y, double xscale, double yscale, int flags, loa_abstract_t*, nil_clip_t*, cairo_t*);

loa_abstract_t *loa_abstract_abydos_create(void *userdata, int width, int height);
void loa_abstract_abydos_free(void *userdata, loa_abstract_t*);
void loa_abstract_abydos_get_size(void *userdata, loa_abstract_t*, int *width, int *height);
loa_abstract_t *loa_abstract_abydos_copy(void *userdata, loa_abstract_t*, int deep);
void loa_abstract_abydos_fill(void *userdata, loa_abstract_t*, double r, double g, double b, double a, nil_clip_t*);
void loa_abstract_abydos_paste(void *userdata, loa_abstract_t *dst, double x, double y, double xscale, double yscale, int tile, loa_abstract_t *src, nil_clip_t*);

#endif /* loa_abstract_abydos_h */
