/*
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <abydos-plugin.h>
#include <debug.h>
#include <libavcodec/avcodec.h>
#include <libavutil/mathematics.h>
#include <stdlib.h>
#include <string.h>

#include "ffmpegdec.h"

/*

  AV_CODEC_ID_MPEG1VIDEO
  AV_CODEC_ID_MPEG2VIDEO      preferred ID for MPEG-1/2 video decoding
  AV_CODEC_ID_H261
  AV_CODEC_ID_H263
  AV_CODEC_ID_RV10
  AV_CODEC_ID_RV20
  AV_CODEC_ID_MJPEG
  AV_CODEC_ID_MJPEGB
  AV_CODEC_ID_LJPEG
  AV_CODEC_ID_SP5X
  AV_CODEC_ID_JPEGLS
  AV_CODEC_ID_MPEG4
  AV_CODEC_ID_RAWVIDEO
  AV_CODEC_ID_MSMPEG4V1
  AV_CODEC_ID_MSMPEG4V2
  AV_CODEC_ID_MSMPEG4V3
  AV_CODEC_ID_WMV1
  AV_CODEC_ID_WMV2
  AV_CODEC_ID_H263P
  AV_CODEC_ID_H263I
  AV_CODEC_ID_FLV1
  AV_CODEC_ID_SVQ1
  AV_CODEC_ID_SVQ3
  AV_CODEC_ID_DVVIDEO
  AV_CODEC_ID_HUFFYUV
  AV_CODEC_ID_CYUV
  AV_CODEC_ID_H264
  AV_CODEC_ID_INDEO3
  AV_CODEC_ID_VP3
  AV_CODEC_ID_THEORA
  AV_CODEC_ID_ASV1
  AV_CODEC_ID_ASV2
  AV_CODEC_ID_FFV1
  AV_CODEC_ID_4XM
  AV_CODEC_ID_VCR1
  AV_CODEC_ID_CLJR
  AV_CODEC_ID_MDEC
  AV_CODEC_ID_ROQ
  AV_CODEC_ID_INTERPLAY_VIDEO
  AV_CODEC_ID_XAN_WC3
  AV_CODEC_ID_XAN_WC4
  AV_CODEC_ID_RPZA
  AV_CODEC_ID_CINEPAK
  AV_CODEC_ID_WS_VQA
  AV_CODEC_ID_MSRLE
  AV_CODEC_ID_MSVIDEO1
  AV_CODEC_ID_IDCIN
  AV_CODEC_ID_8BPS
  AV_CODEC_ID_SMC
  AV_CODEC_ID_FLIC
  AV_CODEC_ID_TRUEMOTION1
  AV_CODEC_ID_VMDVIDEO
  AV_CODEC_ID_MSZH
  AV_CODEC_ID_ZLIB
  AV_CODEC_ID_QTRLE
  AV_CODEC_ID_TSCC
  AV_CODEC_ID_ULTI
  AV_CODEC_ID_QDRAW
  AV_CODEC_ID_VIXL
  AV_CODEC_ID_QPEG
  AV_CODEC_ID_PNG
  AV_CODEC_ID_PPM
  AV_CODEC_ID_PBM
  AV_CODEC_ID_PGM
  AV_CODEC_ID_PGMYUV
  AV_CODEC_ID_PAM
  AV_CODEC_ID_FFVHUFF
  AV_CODEC_ID_RV30
  AV_CODEC_ID_RV40
  AV_CODEC_ID_VC1
  AV_CODEC_ID_WMV3
  AV_CODEC_ID_LOCO
  AV_CODEC_ID_WNV1
  AV_CODEC_ID_AASC
  AV_CODEC_ID_INDEO2
  AV_CODEC_ID_FRAPS
  AV_CODEC_ID_TRUEMOTION2
  AV_CODEC_ID_BMP
  AV_CODEC_ID_CSCD
  AV_CODEC_ID_MMVIDEO
  AV_CODEC_ID_ZMBV
  AV_CODEC_ID_AVS
  AV_CODEC_ID_SMACKVIDEO
  AV_CODEC_ID_NUV
  AV_CODEC_ID_KMVC
  AV_CODEC_ID_FLASHSV
  AV_CODEC_ID_CAVS
  AV_CODEC_ID_JPEG2000
  AV_CODEC_ID_VMNC
  AV_CODEC_ID_VP5
  AV_CODEC_ID_VP6
  AV_CODEC_ID_VP6F
  AV_CODEC_ID_TARGA
  AV_CODEC_ID_DSICINVIDEO
  AV_CODEC_ID_TIERTEXSEQVIDEO
  AV_CODEC_ID_TIFF
  AV_CODEC_ID_GIF
  AV_CODEC_ID_DXA
  AV_CODEC_ID_DNXHD
  AV_CODEC_ID_THP
  AV_CODEC_ID_SGI
  AV_CODEC_ID_C93
  AV_CODEC_ID_BETHSOFTVID
  AV_CODEC_ID_PTX
  AV_CODEC_ID_TXD
  AV_CODEC_ID_VP6A
  AV_CODEC_ID_AMV
  AV_CODEC_ID_VB
  AV_CODEC_ID_PCX
  AV_CODEC_ID_SUNRAST
  AV_CODEC_ID_INDEO4
  AV_CODEC_ID_INDEO5
  AV_CODEC_ID_MIMIC
  AV_CODEC_ID_RL2
  AV_CODEC_ID_ESCAPE124
  AV_CODEC_ID_DIRAC
  AV_CODEC_ID_BFI
  AV_CODEC_ID_CMV
  AV_CODEC_ID_MOTIONPIXELS
  AV_CODEC_ID_TGV
  AV_CODEC_ID_TGQ
  AV_CODEC_ID_TQI
  AV_CODEC_ID_AURA
  AV_CODEC_ID_AURA2
  AV_CODEC_ID_V210X
  AV_CODEC_ID_TMV
  AV_CODEC_ID_V210
  AV_CODEC_ID_DPX
  AV_CODEC_ID_MAD
  AV_CODEC_ID_FRWU
  AV_CODEC_ID_FLASHSV2
  AV_CODEC_ID_CDGRAPHICS
  AV_CODEC_ID_R210
  AV_CODEC_ID_ANM
  AV_CODEC_ID_BINKVIDEO
  AV_CODEC_ID_IFF_ILBM
  AV_CODEC_ID_KGV1
  AV_CODEC_ID_YOP
  AV_CODEC_ID_VP8
  AV_CODEC_ID_PICTOR
  AV_CODEC_ID_ANSI
  AV_CODEC_ID_A64_MULTI
  AV_CODEC_ID_A64_MULTI5
  AV_CODEC_ID_R10K
  AV_CODEC_ID_MXPEG
  AV_CODEC_ID_LAGARITH
  AV_CODEC_ID_PRORES
  AV_CODEC_ID_JV
  AV_CODEC_ID_DFA
  AV_CODEC_ID_WMV3IMAGE
  AV_CODEC_ID_VC1IMAGE
  AV_CODEC_ID_UTVIDEO
  AV_CODEC_ID_BMV_VIDEO
  AV_CODEC_ID_VBLE
  AV_CODEC_ID_DXTORY
  AV_CODEC_ID_V410
  AV_CODEC_ID_XWD
  AV_CODEC_ID_CDXL
  AV_CODEC_ID_XBM
  AV_CODEC_ID_ZEROCODEC
  AV_CODEC_ID_MSS1
  AV_CODEC_ID_MSA1
  AV_CODEC_ID_TSCC2
  AV_CODEC_ID_MTS2
  AV_CODEC_ID_CLLC
  AV_CODEC_ID_MSS2
  AV_CODEC_ID_VP9
  AV_CODEC_ID_AIC
  AV_CODEC_ID_ESCAPE130
  AV_CODEC_ID_G2M
  AV_CODEC_ID_WEBP
  AV_CODEC_ID_HNM4_VIDEO
  AV_CODEC_ID_HEVC
  AV_CODEC_ID_FIC
  AV_CODEC_ID_ALIAS_PIX
  AV_CODEC_ID_BRENDER_PIX
  AV_CODEC_ID_PAF_VIDEO
  AV_CODEC_ID_EXR
  AV_CODEC_ID_VP7
  AV_CODEC_ID_SANM
  AV_CODEC_ID_SGIRLE
  AV_CODEC_ID_MVC1
  AV_CODEC_ID_MVC2
  AV_CODEC_ID_HQX
  AV_CODEC_ID_TDSC
  AV_CODEC_ID_HQ_HQA
  AV_CODEC_ID_HAP
  AV_CODEC_ID_DDS
  AV_CODEC_ID_DXV
  AV_CODEC_ID_SCREENPRESSO
  AV_CODEC_ID_RSCC
  AV_CODEC_ID_AVS2
  AV_CODEC_ID_Y41P
  AV_CODEC_ID_AVRP
  AV_CODEC_ID_012V
  AV_CODEC_ID_AVUI
  AV_CODEC_ID_AYUV
  AV_CODEC_ID_TARGA_Y216
  AV_CODEC_ID_V308
  AV_CODEC_ID_V408
  AV_CODEC_ID_YUV4
  AV_CODEC_ID_AVRN
  AV_CODEC_ID_CPIA
  AV_CODEC_ID_XFACE
  AV_CODEC_ID_SNOW
  AV_CODEC_ID_SMVJPEG
  AV_CODEC_ID_APNG
  AV_CODEC_ID_DAALA
  AV_CODEC_ID_CFHD
  AV_CODEC_ID_TRUEMOTION2RT
  AV_CODEC_ID_M101
  AV_CODEC_ID_MAGICYUV
  AV_CODEC_ID_SHEERVIDEO
  AV_CODEC_ID_YLC
  AV_CODEC_ID_PSD
  AV_CODEC_ID_PIXLET
  AV_CODEC_ID_SPEEDHQ
  AV_CODEC_ID_FMVC
  AV_CODEC_ID_SCPR
  AV_CODEC_ID_CLEARVIDEO
  AV_CODEC_ID_XPM
  AV_CODEC_ID_AV1
  AV_CODEC_ID_BITPACKED
  AV_CODEC_ID_MSCC
  AV_CODEC_ID_SRGC
  AV_CODEC_ID_SVG
  AV_CODEC_ID_GDV
  AV_CODEC_ID_FITS
  AV_CODEC_ID_IMM4
  AV_CODEC_ID_PROSUMER
  AV_CODEC_ID_MWSC
  AV_CODEC_ID_WCMV
  AV_CODEC_ID_RASC
  AV_CODEC_ID_HYMT
  AV_CODEC_ID_ARBC
  AV_CODEC_ID_AGM
  AV_CODEC_ID_LSCR
  AV_CODEC_ID_VP4
  AV_CODEC_ID_IMM5
  AV_CODEC_ID_MVDV
  AV_CODEC_ID_MVHA
  AV_CODEC_ID_CDTOONS

  AV_CODEC_ID_TTF
  AV_CODEC_ID_SCTE_35         Contain timestamp estimated through PCR of program stream.

  AV_CODEC_ID_EPG
  AV_CODEC_ID_BINTEXT
  AV_CODEC_ID_XBIN
  AV_CODEC_ID_IDF
  AV_CODEC_ID_OTF
  AV_CODEC_ID_SMPTE_KLV
  AV_CODEC_ID_DVD_NAV
  AV_CODEC_ID_TIMED_ID3
  AV_CODEC_ID_BIN_DATA

*/

#ifndef CLAMP
# define CLAMP(a,low,high) ((a)>(high)?(high):((a)<(low)?(low):(a)))
#endif

static cairo_surface_t *
_surface_from_frame(int format, AVFrame *frame, int width, int height)
{
    cairo_surface_t *surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, width, height);
    const uint8_t *src = frame->data[0];
    const uint8_t *src2;
    const uint8_t *src3;
    int src_skip = frame->linesize[0] - width*4;
    int src2_skip;
    int src3_skip;
    uint32_t *dst = (uint32_t*)cairo_image_surface_get_data(surface);
    int dst_skip = cairo_image_surface_get_stride(surface)/sizeof(uint32_t) - width;
    int y;
    switch(format){
    case AV_PIX_FMT_RGBA:
        while(height-->0){
            int x;
            for(x=0;x<width;++x){
                int alpha = src[3];
                int r = alpha*src[0]/255;
                int g = alpha*src[1]/255;
                int b = alpha*src[2]/255;
                *dst++ = r<<16 | g<<8 | b<<0 | alpha<<24;
                src += 4;
            }
            src += src_skip;
            dst += dst_skip;
        }
        break;
    case AV_PIX_FMT_YUV411P:
        src_skip = frame->linesize[0] - width;
        src2 = frame->data[1];
        src3 = frame->data[2];
        src2_skip = frame->linesize[1] - width/4;
        src3_skip = frame->linesize[2] - width/4;
        while(height-->0){
            int x;
            for(x=0;x<width;++x){
                int y = src[0];
                int u = src2[0];
                int v = src3[0];
                int r = y + (1.370705 * (v-128));
                int g = y - (0.698001 * (v-128)) - (0.337633 * (u-128));
                int b = y + (1.732446 * (u-128));
                r = CLAMP(r,0,255);
                g = CLAMP(g,0,255);
                b = CLAMP(b,0,255);
                *dst++ = r<<16 | g<<8 | b<<0 | 255<<24;
                ++src;
                if(3==(x&3)){
                    ++src2;
                    ++src3;
                }
            }
            src += src_skip;
            src2 += src2_skip;
            src3 += src3_skip;
            dst += dst_skip;
        }
        break;
    case AV_PIX_FMT_YUV420P:
    case AV_PIX_FMT_YUVJ420P:
        src_skip = frame->linesize[0] - width;
        src2 = frame->data[1];
        src3 = frame->data[2];
        src2_skip = frame->linesize[1] - width/2;
        src3_skip = frame->linesize[2] - width/2;
        for(y=0;y<height;++y){
            int x;
            for(x=0;x<width;++x){
                int y = src[0];
                int u = src2[0];
                int v = src3[0];
                int r = y + (1.370705 * (v-128));
                int g = y - (0.698001 * (v-128)) - (0.337633 * (u-128));
                int b = y + (1.732446 * (u-128));
                r = CLAMP(r,0,255);
                g = CLAMP(g,0,255);
                b = CLAMP(b,0,255);
                *dst++ = r<<16 | g<<8 | b<<0 | 255<<24;
                ++src;
                if(x&1){
                    ++src2;
                    ++src3;
                }
            }
            src += src_skip;
            src2 += src2_skip;
            src3 += src3_skip;
            if(!(y&1)){
                src2 -= frame->linesize[1];
                src3 -= frame->linesize[2];
            }
            dst += dst_skip;
        }
        break;
    case AV_PIX_FMT_YUV422P:
        src_skip = frame->linesize[0] - width;
        src2 = frame->data[1];
        src3 = frame->data[2];
        src2_skip = frame->linesize[1] - width/2;
        src3_skip = frame->linesize[2] - width/2;
        while(height-->0){
            int x;
            for(x=0;x<width;++x){
                int y = src[0];
                int u = src2[0];
                int v = src3[0];
                int r = y + (1.370705 * (v-128));
                int g = y - (0.698001 * (v-128)) - (0.337633 * (u-128));
                int b = y + (1.732446 * (u-128));
                r = CLAMP(r,0,255);
                g = CLAMP(g,0,255);
                b = CLAMP(b,0,255);
                *dst++ = r<<16 | g<<8 | b<<0 | 255<<24;
                ++src;
                if(x&1){
                    ++src2;
                    ++src3;
                }
            }
            src += src_skip;
            src2 += src2_skip;
            src3 += src3_skip;
            dst += dst_skip;
        }
        break;
    default:
        /*
        while(height-->0){
            int x;
            for(x=0;x<width;++x){
                int alpha = src[3];
                int r = alpha*src[2]/255;
                int g = alpha*src[1]/255;
                int b = alpha*src[0]/255;
                *dst++ = r<<16 | g<<8 | b<<0 | alpha<<24;
                src += 4;
            }
            src += src_skip;
            dst += dst_skip;
        }
        */
        break;
    }
    cairo_surface_mark_dirty(surface);
    return surface;
}


cairo_surface_t *
ffmpeg_decode(const char *data, ssize_t len, enum AVCodecID id)
{
    cairo_surface_t *surface = NULL;
    AVCodec *codec = NULL;
    AVCodecContext *ctx = NULL;
    AVFrame *frame = NULL;
    AVPacket packet;
    int ret;

    codec = avcodec_find_decoder(id);
    if(!codec)
        return NULL;
    ctx = avcodec_alloc_context3(codec);
    if(avcodec_open2(ctx, codec, NULL) < 0)
        goto out;

    frame = av_frame_alloc();

    av_init_packet(&packet);
    packet.data = (unsigned char*)data;
    packet.size = len;


    if(avcodec_send_packet(ctx, &packet) < 0) {
        DEBUG_p("error during decoding");
        goto out;
    }
#ifdef USE_DEBUG
    switch(ctx->sw_pix_fmt /* h->ctx->pix_fmt */){
    case AV_PIX_FMT_RGBA:
        DEBUG_p("RGBA");
        break;
    case AV_PIX_FMT_YUV411P:
        DEBUG_p("YUV411P");
        break;
    case AV_PIX_FMT_YUV420P:
        DEBUG_p("YUV420P");
        break;
    case AV_PIX_FMT_YUVJ420P:
        DEBUG_p("YUVJ420P");
        break;
    case AV_PIX_FMT_YUV422P:
        DEBUG_p("YUV422P");
        break;
    default:
        DEBUG_p("unhandled pixel format");
        break;
    }
#endif
    //while (ret >= 0){
        ret = avcodec_receive_frame(ctx, frame);
        if (ret == AVERROR(EAGAIN) || ret == AVERROR_EOF){
            /* frame found */
            goto out;
        }else if (ret < 0) {
            DEBUG_p("error during decoding");
            goto out;
        }
        surface = _surface_from_frame(ctx->sw_pix_fmt,frame,frame->width,frame->height);
#if 0
        ret = av_hwframe_transfer_data(sw_frame, frame, 0);
        if (ret < 0) {
            DEBUG_p("error transferring the data to system memory");
            goto out;
        }
#endif
    //}

out:
    if(frame)
        av_frame_free(&frame);
    if(ctx)
        avcodec_free_context(&ctx);
    //if(codec)
    //    avcodec_free(&codec);
    return surface;
}
