/*
 * Copyright (C) 2019  Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include <debug.h>
#include <math.h>
#include <stddef.h>

//#include "uemf.h"
//#include "uemf_safe.h"
#include "uwmf.h"
#include "wmfcairo.h"


//#define UM2PX (27.0/10)
#define UM2PX (0.00370370370369)
#define SCALE (0.07988017973)


#if 0
static uint32_t lu_crc32(const uint8_t *message, uint32_t len){
   uint32_t i, j;
   uint32_t crc, mask;

   crc = 0xFFFFFFFF;
   for(i=0;i<len;i++){     // over all bytes
      crc = crc ^ *message++;
      for (j = 0; j < 8; j++) { // over all bits
         mask = -(crc & 1);
         crc = (crc >> 1) ^ (0xEDB88320 & mask);
      }
   }
   return ~crc;
}
#endif

/* --- utility drawing functions --- */

typedef struct{
    uint16_t count;
    const U_POINT16 *pt;
}polyline_t;


static void
_cairo_draw_poly(cairo_t *cr, polyline_t *pl)
{
    int i;
    cairo_move_to(cr,pl->pt[0].x,pl->pt[0].y);
    for(i=1;i<pl->count;++i)
        cairo_line_to(cr,pl->pt[i].x,pl->pt[i].y);
}


#if 0
static void
_cairo_set_fill_style(cairo_t *cr, wmfDC *dc)
{
    wmfBrush *brush = WMF_DC_BRUSH(dc);
    switch(WMF_DC_POLYFILL(dc)){
    case WINDING:
        cairo_set_fill_rule(cr,CAIRO_FILL_RULE_WINDING);
        break;
    case ALTERNATE:
    default:
        cairo_set_fill_rule(cr,CAIRO_FILL_RULE_EVEN_ODD);
        break;
    }
    _cairo_set_rgb(cr,WMF_BRUSH_COLOR(brush));
}
#endif


static void
_cairo_set_brush_style(cairo_t *cr, U_WLOGBRUSH *brush)
{
    DEBUG_p("color: %d,%d,%d",brush->Color.Red,brush->Color.Green,brush->Color.Blue);
    cairo_set_source_rgb(cr,(double)brush->Color.Red/255,(double)brush->Color.Green/255,(double)brush->Color.Blue/255);
}


#if 0
static void
_cairo_set_stroke_style(cairo_t *cr, wmfDC *dc)
{
    wmfPen* pen = WMF_DC_PEN(dc);
    float pen_width = (float)(WMF_PEN_WIDTH(pen)+WMF_PEN_HEIGHT(pen))/2;
    int pen_style = WMF_PEN_STYLE(pen);

    if(PS_NULL==pen_style){
        DEBUG_p("penstyle null");
    }

    cairo_set_line_width(cr,pen_width);

    switch (WMF_PEN_ENDCAP(pen)){
    case PS_ENDCAP_SQUARE:
        cairo_set_line_cap(cr,CAIRO_LINE_CAP_SQUARE);
        break;
    case PS_ENDCAP_ROUND:
        cairo_set_line_cap(cr,CAIRO_LINE_CAP_ROUND);
        break;
    case PS_ENDCAP_FLAT:
    default:
        cairo_set_line_cap(cr,CAIRO_LINE_CAP_BUTT);
        break;
    }

    switch (WMF_PEN_JOIN(pen)){
    case PS_JOIN_BEVEL:
        cairo_set_line_join(cr,CAIRO_LINE_JOIN_BEVEL);
        break;
    case PS_JOIN_ROUND:
        cairo_set_line_join(cr,CAIRO_LINE_JOIN_ROUND);
        break;
    case PS_JOIN_MITER:
    default:
        cairo_set_line_join(cr,CAIRO_LINE_JOIN_MITER);
        break;
    }

#if 0
    switch (pen_style){
    case PS_DASH: /* DASH_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f; ", pen_width*10,pen_width*10);
        break;

    case PS_ALTERNATE:
    case PS_DOT: /* DOTTED_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f; ", pen_width,pen_width*2);
        break;

    case PS_DASHDOT: /* DASH_DOT_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f %f %f; ", pen_width*10,pen_width*2,pen_width,pen_width*2);
        break;

    case PS_DASHDOTDOT: /* DASH_2_DOTS_LINE */
        wmf_stream_printf (API,out,"stroke-dasharray:%f %f %f %f %f %f; ", pen_width*10,pen_width*2,pen_width,pen_width*2,pen_width,pen_width*2);
        break;

    case PS_INSIDEFRAME: /* There is nothing to do in this case... */
    case PS_SOLID:
    default:
        wmf_stream_printf (API,out,"stroke-dasharray:none; ");
        break;
    }
#endif

    _cairo_set_rgb(cr,WMF_PEN_COLOR(pen));
}
#endif


#if 0
static void
_cairo_fill_and_stroke(cairo_t *cr, wmfDC *dc)
{
    _cairo_set_fill_style(cr,dc);
    cairo_fill_preserve(cr);
    _cairo_set_stroke_style(cr,dc);
    cairo_stroke(cr);
}
#endif

/* --- WMF implementation --- */

typedef struct{
    cairo_t *cr;
    U_WLOGBRUSH *brush;
    void *pen;
}state_t;


static void
on_createbrushindirect(U_WMRCREATEBRUSHINDIRECT *wmr, state_t *state)
{
    U_LOGBRUSH *brush;
    U_WMRCREATEBRUSHINDIRECT_get((const char*)wmr,(const char**)&brush);
    //DEBUG_p("color: %d,%d,%d",wmr->brush.Color.Red,wmr->brush.Color.Green,wmr->brush.Color.Blue);
    state->brush = &wmr->brush;
}


static void
on_createpenindirect(U_WMRCREATEPENINDIRECT *emr, state_t *state)
{
    DEBUG_p("begin");
}


static void on_deleteobject(){
    DEBUG_p("begin");
}


static void
on_eof(U_WMREOF *wmr, state_t *state)
{
    DEBUG_p("begin");
}


static void
on_escape()
{
    DEBUG_p("begin");
}


static void on_lineto(U_WMRLINETO *wmr, state_t *state){
    U_POINT16 pt;
    DEBUG_p("begin");
    U_WMRLINETO_get((char*)wmr,&pt);
    cairo_line_to(state->cr,pt.x,pt.y);
}


static void
on_polygon(U_WMRPOLYGON *wmr, state_t *state)
{
    DEBUG_p("begin");
    polyline_t pl;
#if 1
    pl.count = wmr->nPoints;
    pl.pt = wmr->aPoints;
#else
    U_WMRPOLYGON_get(wmr,&pl.count,(const char**)&pl.pt);
#endif

    if(pl.count <= 2)
        return;

#if 0
    if(TO_FILL(poly_line) || TO_DRAW(poly_line)){
        _cairo_draw_poly(cr,poly_line);
        _cairo_fill_and_stroke(cr,poly_line->dc);
    }
#else
    _cairo_draw_poly(state->cr,&pl);
    _cairo_set_brush_style(state->cr,state->brush);
    cairo_fill(state->cr);
#endif
}


static void
on_polyline()
{
    DEBUG_p("begin");
}


static void
on_polypolygon(U_WMRPOLYPOLYGON *wmr, state_t *state)
{
    uint16_t polygon_count;
    const uint16_t *pt_count;
    const U_POINT16 *pt;
    int i;
    DEBUG_p("begin");
    U_WMRPOLYPOLYGON_get((char*)wmr,&polygon_count,&pt_count,(const char**)&pt);
    for(i=0;i<polygon_count;++i){
        polyline_t pl;
        pl.count = *pt_count;
        pl.pt = pt;
        _cairo_draw_poly(state->cr,&pl);
        pt += *pt_count++;
    }
    _cairo_set_brush_style(state->cr,state->brush);
    cairo_fill(state->cr);
}


static void
on_selectobject()
{
    //DEBUG_p("begin");
}

static void
on_setbkmode(U_WMRSETBKMODE *wmr, state_t *state)
{
    DEBUG_p("begin");
}


static void
on_setmapmode(U_WMRSETMAPMODE *wmr, state_t *state)
{
    DEBUG_p("begin");
}


static void
on_setpolyfillmode()
{
    DEBUG_p("begin");
}


static void on_setrop2(){
    DEBUG_p("begin");
}


static void on_settextalign(){
    DEBUG_p("begin");
}


static void on_settextcolor(U_WMRSETTEXTCOLOR *wmr, state_t *state){
    DEBUG_p("begin");
    //cairo_set_source_rgb(cr,(double)wmr->Color.Red/255,(double)wmr->Color.Green/255,(double)wmr->Color.Blue/255);
}


static void
on_setwindowext(U_WMRSETWINDOWEXT *wmr, state_t *state)
{
    DEBUG_p("begin");
}


static void
on_setwindoworg(U_WMRSETWINDOWORG *wmr, state_t *state)
{
    DEBUG_p("begin");
}

/* --- */

#if 0
static void on_polybezierto(U_EMRPOLYBEZIERTO *emr,cairo_t *cr){
    //DEBUG_p("points: %d",emr->cptl);
    cairo_curve_to(cr,emr->aptl[0].x,emr->aptl[0].y,emr->aptl[1].x,emr->aptl[1].y,emr->aptl[2].x,emr->aptl[2].y);
}
#endif
/*
static void on_movetoex(U_WMRMOVETOEX *emr,cairo_t *cr){
    cairo_move_to(cr,emr->ptl.x,emr->ptl.y);
}
*/
/*
static void on_fillpath(U_WMRFILLPATH *emr,cairo_t *cr){
    DEBUG_p("bounds: %dx%d - %dx%d",emr->rclBounds.left,emr->rclBounds.top,emr->rclBounds.right,emr->rclBounds.bottom);
    cairo_fill(cr);
}
*/


static ssize_t
handle_rec(U_METARECORD *wmr, state_t *state)
{
#if 0
    uint32_t crc;
#if __BYTE_ORDER == __BIG_ENDIAN
    {
        char *swapbuf = malloc(size);
        memcpy(swapbuf,record,size);
        U_emf_endian(swapbuf,size,1);
        crc = lu_crc32(swapbuf,size);
        free(swapbuf);
    }
#else
    crc = lu_crc32((const uint8_t*)emr,emr->nSize);
#endif
    //DEBUG_p("crc: %d",crc);
#endif

    switch(wmr->iType){
    case U_WMR_EOF:
        on_eof((U_WMREOF*)wmr,state);
        return 0;
    case U_WMR_SETBKMODE:
        on_setbkmode((U_WMRSETBKMODE*)wmr,state);
        break;
    case U_WMR_SETMAPMODE:
        on_setmapmode((U_WMRSETMAPMODE*)wmr,state);
        break;
    case U_WMR_SETROP2:
        on_setrop2(wmr,state);
        break;

    /* 0x05 */

    case U_WMR_SETPOLYFILLMODE:
        on_setpolyfillmode((U_WMRSETPOLYFILLMODE*)wmr,state);
        break;

    /* 0x07 - 0x08 */

    case U_WMR_SETTEXTCOLOR:
        on_settextcolor((U_WMRSETTEXTCOLOR*)wmr,state);
        break;

    /* 0x0a */

    case U_WMR_SETWINDOWORG:
        on_setwindoworg((U_WMRSETWINDOWORG*)wmr,state);
        break;
    case U_WMR_SETWINDOWEXT:
        on_setwindowext((U_WMRSETWINDOWEXT*)wmr,state);
        break;

    /* 0x0d - 0x12 */

    case U_WMR_LINETO:
        on_lineto((U_WMRLINETO*)wmr,state);
        break;

    /* 0x14 - 0x23 */

    case U_WMR_POLYGON:
        on_polygon((U_WMRPOLYGON*)wmr,state);
        break;
    case U_WMR_POLYLINE:
        on_polyline(wmr,state);
        break;
    case U_WMR_ESCAPE:
        on_escape(wmr,state);
        break;

    /* 0x27 - 0x2c */

    case U_WMR_SELECTOBJECT:
        on_selectobject(wmr,state);
        break;
    case U_WMR_SETTEXTALIGN:
        on_settextalign(wmr,state);
        break;

    /* 0x2f - 0x37 */

    case U_WMR_POLYPOLYGON:
        on_polypolygon((U_WMRPOLYPOLYGON*)wmr,state);
        break;

    /* 0x39 - 0xef */

    case U_WMR_DELETEOBJECT:
        on_deleteobject(wmr,state);
        break;

    /* 0xf1 - 0xf9 */

    case U_WMR_CREATEPENINDIRECT:
        on_createpenindirect((U_WMRCREATEPENINDIRECT*)wmr,state);
        break;

    /* 0xfb */

    case U_WMR_CREATEBRUSHINDIRECT:
        on_createbrushindirect((U_WMRCREATEBRUSHINDIRECT*)wmr,state);
        break;

    /* 0xfd - */

/*
    case U_WMR_MOVETOEX:
        on_movetoex((U_EMRMOVETOEX*)wmr,cr);
        break;
    case U_WMR_MODIFYWORLDTRANSFORM:
        on_modifyworldtransform((U_EMRMODIFYWORLDTRANSFORM*)wmr,cr);
        break;
    case U_WMR_BEGINPATH:
        on_beginpath((U_EMRBEGINPATH*)wmr,cr);
        break;
    case U_WMR_ENDPATH:
        on_endpath((U_EMRENDPATH*)wmr,cr);
        break;
    case U_WMR_CLOSEFIGURE:
        on_closefigure(wmr,cr);
        break;
    case U_WMR_FILLPATH:
        on_fillpath((U_EMRFILLPATH*)wmr,cr);
        break;
    case U_WMR_COMMENT:
        on_comment(wmr,cr);
        break;
*/
    default:
        DEBUG_p("unhandled type: 0x%02x",wmr->iType);
        break;
    }
    return 1;
}


int wmf_cairo_scan(const char* p,size_t len,int *width,int *height)
{
    const char *end = p+len;
    U_WMRPLACEABLE placeable;
    U_WMRHEADER header;
    int size;
    size = wmfheader_get(p,end,&placeable,&header);
    if(size<=0)
        return -1;
    *width = ceil(SCALE*placeable.Dst.right);
    *height = ceil(SCALE*placeable.Dst.bottom);
    return 0;
}


int wmf_cairo_render(const char *p, size_t len, cairo_t *cr)
{
    state_t state;
    const char *end = p+len;
    {
        U_WMRPLACEABLE placeable;
        U_WMRHEADER header;
        int headersize = wmfheader_get(p,end,&placeable,&header);
        if(headersize<=0)
            return -1;
        p += headersize;
    }
    cairo_scale(cr,SCALE,SCALE);
    state.cr = cr;
    state.brush = NULL;
    state.pen = NULL;
    while(p<end){
        size_t size = U_WMRRECSAFE_get(p,end);
        U_METARECORD *wmr;
        ssize_t result;
        if(size<=0){
            DEBUG_p("corrupt record");
            return -1;
        }
        wmr = (U_METARECORD*)p;
        result = handle_rec(wmr, &state);
        if(result < 0)
            return -1;
        if(!result)
            return 0;

        p += size;
    }
    return 0;
}
