/*
 * Copyright (C) 2017  Teemu Suutari
 * Copyright (C) 2020  Magnus Bergman <magnus@snisurset.net>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * * Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * * Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 */
#ifndef PPDECOMPRESSOR_HPP
#define PPDECOMPRESSOR_HPP

#include "dec.h"

class PPDecompressor : public Decompressor, public XPKDecompressor
{
private:
	class PPState : public XPKDecompressor::State
	{
	public:
		PPState(uint32_t mode);
		virtual ~PPState();

		uint32_t _cachedMode;
	};

public:
	PPDecompressor(const Buffer &packedData,bool exactSizeKnown,bool verify);
	PPDecompressor(const Buffer &packedData,std::unique_ptr<XPKDecompressor::State> &state);
	virtual ~PPDecompressor(){}

	//virtual size_t getPackedSize() const noexcept override final;
	//virtual size_t getRawSize() const noexcept override final;

	virtual void decompressImpl(Buffer &rawData,const Buffer &previousData,bool verify) override final;

private:
	size_t		_dataStart=0;
	size_t		_rawSize=0;
	uint8_t		_startShift=0;
	uint8_t		_modeTable[4];
	bool		_isXPK=false;
};


#endif
