/*
 * Copyright (C) 2015-2020  Magnus Bergman
 *
 * Author: Magnus Bergman <magnus@snisurset.net>
 *
 * This library is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or (at
 * your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http:/www.gnu.org/licenses/>
 */
#include "list.h"

#ifdef LIST_USE_MOWGLI

/* FIXME!! */

#elif defined(LIST_USE_GLIB)

/* Nothing */

#else

#include "debug.h"
#include "memutil.h"


static List* list_new_node(void* data){
    List* node = u_new(List,1);
    node->data = data;
    node->next = NULL;
    node->prev = NULL;
    return node;
}


void list_free(List* list){
    while(list){
        List* tmp = list;
        list = list->next;
        u_free(tmp);
    }
}


void list_free_full(List* list,ListNotifyFunc func){
    while(list){
        List* tmp = list;
        list = list->next;
        func(tmp->data);
        u_free(tmp);
    }
}


List* list_insert_before(List* list,List* node,void* data){
    List* newnode;
    newnode = list_new_node(data);
    if(!list)
        return newnode;
    if(!node)
        node = list;
    newnode->next = node;
    newnode->prev = node->prev;
    node->prev = newnode;
    if(newnode->prev){
        newnode->prev->next=newnode;
        return list;
    }else{
        /* node is the first node in list */
        return newnode;
    }
}


List* list_insert_after(List* list,List* node,void* data){
    List* newnode;
    newnode = list_new_node(data);
    if(!list)
        return newnode;
    if(!node){
        DEBUG_p("appending node");
        node = list;
        while(node->next)
            node = node->next;
    }
    newnode->next = node->next;
    newnode->prev = node;
    node->next = newnode;
    if(newnode->next)
        newnode->next->prev = newnode;
    return list;
}


List* list_insert_sorted(List* list,void* data,ListCompFunc func){
    List* iter;
    if(!list)
        return list_new_node(data);
    for(iter=list;;iter=iter->next){
        if(func(data,iter->data)<0)
            return list_insert_before(list,iter,data);
        if(!iter->next)
            return list_insert_after(list,iter,data);
    }
}



List* list_find_custom(List* list,ListCompFunc func,void* data){
    List* iter;
    for(iter=list;iter;iter=iter->next)
        if(!func(iter->data,data))
            return iter;
    return NULL;
}


List* list_find_data(List* list,void* data){
    List* iter;
    for(iter=list;iter;iter=iter->next)
        if(iter->data==data)
            return iter;
    return NULL;
}


void list_foreach(List* list,ListForeachFunc func,void* userdata){
    for(;list;list=list->next)
        func(list->data,userdata);
}



List* list_remove(List* list,List* node){
    if(!list || !node)
        return list;
    if(node->next)
        node->next->prev = node->prev;
    if(node->prev){
        node->prev->next = node->next;
    }else{
        /* if there is no previous node, then list==node */
        list = node->next;
    }
    u_free(node);
    return list;
}


List* list_remove_custom(List* list,ListCompFunc func,void* data){
    List* iter;
    for(iter=list;iter;iter=iter->next)
        if(!func(iter,data))
            return list_remove(list,iter);
    return list;
}


List* list_remove_data(List* list,const void* data){
    List* iter;
    for(iter=list;iter;iter=iter->next)
        if(iter->data==data)
            return list_remove(list,iter);
    return list;
}

#endif /* !USE_GLIB */
